from __future__ import annotations

import asyncio
from abc import ABC
from asyncio import Future
from typing import List, Dict, Any, Generic

from web_foundation import settings
from web_foundation.environment.metrics.dispatcher import MetricsDispatcher
from web_foundation.environment.resources.file_repo.repo import FileRepository
from web_foundation.environment.resources.stores.store import AppStore
from web_foundation.environment.services.service import Service
from web_foundation.kernel.abstract.container import GenericDepContainer
from web_foundation.kernel.configuration import GenericConfig
from web_foundation.kernel.messaging.dispatcher import IDispatcher
from web_foundation.kernel.worker import GenWorker, Worker


class AbstractApp(ABC, Generic[GenericDepContainer]):
    name: str
    workers: Dict[str, Worker]
    container: GenericDepContainer
    dispatcher: IDispatcher

    def __init__(self, container: GenericDepContainer):
        self.container = container
        self.workers = {}
        self.services = {}

    async def _before_app_run(self):
        pass

    async def before_app_run(self):
        pass

    def _add_worker(self, worker: GenWorker):
        pass

    def add_worker(self, worker: GenWorker | List[GenWorker]):
        """
        Set isolate to environment and set isolate debug and name
        :param worker: isolate to apply
        :return: None
        """
        if isinstance(worker, list):
            for w in worker:
                self._add_worker(w)
        else:
            self._add_worker(worker)

    def add_service(self, service: Service):
        self.services.update({service.__class__.__name__: service})

    def perform(self) -> List[Future[Any]]:
        """
        Call performs from isolates and return all isolates Futures
        :return: None
        """
        return [worker.perform() for worker in self.workers.values()]

    async def run(self):
        await self._before_app_run()
        """
        Func to run environment manually (without Runner)
        :return: None
        """
        return await asyncio.wait(self.perform() + self.dispatcher.preform())



import os.path
from enum import Enum
from pathlib import Path
from typing import Any, List

from aiofiles import open
from aiofiles.base import AiofilesContextManager
from loguru import logger

from web_foundation.environment.resources.file_repo.repo import FileRepository, AppFileSections, GenFileSections
from web_foundation import settings
from web_foundation.errors.io.files import SectionNotFound, FileNotExist, NestedFolderDetected, NothingToWrite, \
    OsIOError


class SystemFileRepository(FileRepository):
    _root: Path

    def __init__(self, root: Path, sections: GenFileSections = AppFileSections):
        super(SystemFileRepository, self).__init__(sections)
        self._root = root
        self._sections = sections
        self._create_sections()

    def _create_sections(self):
        for section in self._sections:
            folder = os.path.join(self._root, str(section.value))
            if not self._check_exist(folder, raises=False):
                os.mkdir(folder)

    def _check_exist(self, path: str, raises: bool = True) -> bool:
        exist = os.path.exists(path)
        if not exist and raises:
            raise FileNotExist(message=f"File not exists in path {path}")
        return exist

    def _get_full_path(self, filename: str, section: GenFileSections):
        if section not in self._sections:
            raise SectionNotFound(message=f"Not found {section} in {self._sections}")
        full_path = os.path.join(self._root, str(section.value), filename)
        return full_path

    def _check_full_path(self, filename: str, section: GenFileSections, raises: bool = True):
        full_path = self._get_full_path(filename, section)
        self._check_exist(full_path, raises=raises)
        if os.path.isdir(full_path) and raises:
            raise NestedFolderDetected(message=f"{full_path} is dir")
        return full_path

    # overload
    async def _open_section(self, section: GenFileSections) -> List[str]:
        path = os.path.join(self._root, str(section.value))
        self._check_exist(path)
        filenames = []
        for inner_path in os.listdir(path):
            if os.path.isdir(os.path.join(path, inner_path)):
                continue
            filenames.append(inner_path)
        return filenames

    # overload
    async def stash(self, filename: str, section: GenFileSections, data: Any, raises: bool = True, **kwargs) -> None:
        if not data:
            raise NothingToWrite(message="No data to write")
        try:
            full_path = self._get_full_path(filename, section)  # TODO create dirs if "/" in filename?
            async with open(full_path, **kwargs) as target_file:
                await target_file.write(data)
        except Exception as e:
            if settings.DEBUG:
                logger.debug(f"File write error {e}")
            if raises:
                raise OsIOError(message="Can't write to file", ex=e)

    # overload
    async def take(self, filename: str, section: GenFileSections, raises: bool = True,
                   **kwargs) -> AiofilesContextManager | None:
        try:
            full_path = self._check_full_path(filename, section, raises=raises)
            return open(full_path, **kwargs)
        except Exception as e:
            if settings.DEBUG:
                logger.debug(f"File open error {e}")
            if raises:
                raise OsIOError(message="Can't open file", ex=e)
            return None

    async def get_file_fullpath(self, filename: str, section: Enum) -> str:
        pass



from typing import Any, Dict

from loguru import logger

from web_foundation import settings
from web_foundation.environment.resources.stores.store import AppStore


class InMemoryDictStore(AppStore):
    need_sync = True
    _storage: Dict

    def __init__(self, *args, **kwargs):
        self._storage = {}

    async def _set_item(self, key: str, value: Any):
        if settings.DEBUG:
            if self._storage.get(key):
                logger.debug(f"the key ({key}) will be overwritten")
        self._storage[key] = value

    async def get_item(self, key: str) -> Any:
        ret_val = self._storage.get(key)
        if settings.DEBUG and not ret_val:
            logger.debug(f"""Can"t find key "{key}" in store""")
        return ret_val

    async def size(self):
        return len(self._storage)

    async def get_all(self):
        return self._storage.items()



from enum import Enum
from typing import Any, List, TypeVar, Generic

from aiofiles.base import AiofilesContextManager


class AppFileSections(Enum):
    CONFIG = "config"
    LOGS = "logs"
    PLUGINS = "plugins"
    MIGRATIONS = "migrations"


GenFileSections = TypeVar("GenFileSections", bound=AppFileSections)


class FileLoadContextManager:

    def __aenter__(self):
        raise NotImplementedError

    def __aexit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError


class FileRepository(Generic[GenFileSections]):
    _sections: GenFileSections

    def __init__(self, sections: GenFileSections = AppFileSections):
        self._sections = sections

    async def stash(self, filename: str, section: GenFileSections, data: Any, raises: bool = True, **kwargs):
        raise NotImplementedError

    async def take(self, filename: str, section: GenFileSections, raises: bool = True,
                   **kwargs) -> AiofilesContextManager:
        raise NotImplementedError

    async def _open_section(self, section: GenFileSections) -> List[str]:
        raise NotImplementedError

    async def open_section(self, section: GenFileSections) -> List[str]:
        return await self._open_section(section)

    async def get_file_fullpath(self, filename: str, section: GenFileSections) -> str:
        raise NotImplementedError



from __future__ import annotations

import asyncio
import copy
import os
import socket
from functools import partial
from typing import List, Callable, Any, Type, Generic

import orjson
from sanic import Sanic
from sanic.server.socket import bind_socket

from web_foundation.environment.metrics.mixin import ApiMetricsMixin
from web_foundation.environment.workers.web.ext.addons_loader import AddonsLoader
from web_foundation.environment.workers.web.ext.error_handler import ExtendedErrorHandler
from web_foundation.environment.workers.web.ext.router import ExtRouter
from web_foundation.environment.workers.web.realtime.rt_connection import RtConnection, RtEventCallback
from web_foundation.kernel import GenericIMessage, IMessage
from web_foundation.kernel.abstract.container import GenericDepContainer
from web_foundation.kernel.configuration import ServerConfig
from web_foundation.kernel.worker import Worker


class HttpServer(Worker, ApiMetricsMixin, Generic[GenericDepContainer]):
    config: ServerConfig
    sock: socket.socket
    sanic_app: Sanic
    router: ExtRouter | None
    rt_connections: List[RtConnection]
    api_addons_loader: AddonsLoader | None

    def __init__(self, config: ServerConfig, *args, **kwargs):
        super(HttpServer, self).__init__(*args, **kwargs)
        self.config = config

    def _configure(self, sock: socket.socket = None, ext_router: ExtRouter = None,
                   api_addons_loader: AddonsLoader = None):  # type: ignore
        self.sock = sock if sock else self.create_socket(self.config.host, self.config.port)
        self.router = copy.deepcopy(ext_router)
        self.sanic_app = Sanic(self.name, router=self.router, dumps=orjson.dumps, loads=orjson.loads)
        self.sanic_app.error_handler = ExtendedErrorHandler()
        self.sanic_app.after_server_stop(self.close)
        self.rt_connections = []
        self.api_addons_loader = api_addons_loader

    def _post_configure(self):
        self._set_sanic_confs()
        if self.api_addons_loader and not (self.repo or self.store):
            raise AttributeError(f"You need FileRepository and AppStore to use MiddlewareManager")
        if self.api_addons_loader and issubclass(self.api_addons_loader.__class__, AddonsLoader):
            self.api_addons_loader.store = self.store
            self.api_addons_loader.repo = self.repo
            self.api_addons_loader.reg_channel_middleware()

    def _set_sanic_confs(self):
        self.sanic_app.config.SWAGGER_UI_CONFIGURATION = {
            "docExpansion": 'none'
        }

    async def after_worker_fork(self):
        for service, name in self.container.services():
            service.worker = self

    async def _before_sanic(self, app, loop):
        await self._container.init_resources()
        await self._after_worker_fork()
        if self.api_addons_loader:
            await self.api_addons_loader.discover_middleware()
            await self.api_addons_loader.import_middleware()

        # noinspection PyAsyncCall
        self.sanic_app.add_task(self.channel.listen_consume())

    def _startup(self):
        if self.router and issubclass(self.router.__class__, ExtRouter):
            self.router.apply_routes(self.sanic_app, self.container, self.api_addons_loader)
        # noinspection PyAsyncCall
        self.sanic_app.before_server_start(self._before_sanic)
        # asyncio.run(self._run())
        try:
            self.sanic_app.run(sock=self.sock)
        except KeyboardInterrupt:
            self._close()

    @staticmethod
    def create_socket(host: str, port: int) -> socket.socket:
        sock = bind_socket(host, port)
        sock.set_inheritable(True)
        return sock

    async def broadcasts_rt(self, message: GenericIMessage, resolve_callback: RtEventCallback = None):
        promises = []
        for conn in self.rt_connections:
            if conn.resolve_callback.__name__ == resolve_callback.__name__:
                promises.append(conn.send_after_call(message))
        await asyncio.gather(*promises)

    async def accept_rt_connection(self, conn: RtConnection,
                                   on_disconnect_callback: Callable[[HttpServer, RtConnection], Any] | None = None):
        # conn.debug = self.debug if self.debug else conn.debug
        self.rt_connections.append(conn)

        async def _on_rt_close():
            nonlocal self
            nonlocal conn
            nonlocal on_disconnect_callback
            self.rt_connections.remove(conn)
            if on_disconnect_callback:
                await on_disconnect_callback(self, conn)

        await conn.freeze_request(_on_rt_close)


def create_io_workers(server_cls: Type[HttpServer],
                      config: ServerConfig,
                      router: ExtRouter,
                      api_addons_loader: AddonsLoader = None,
                      workers_num: int = 1,
                      fast: bool = False,
                      **kwargs
                      ) -> List[HttpServer]:
    """
    RESERVE 0 WORKER TO CREATE FRONT SERVICE SERVING
    """
    sock = HttpServer.create_socket(config.host,
                                    config.port)
    workers: List[HttpServer] = []
    w_num = workers_num
    if fast and workers_num != 1:
        raise RuntimeError("You cannot use both fast=True and workers=X")
    if fast:
        try:
            w_num = len(os.sched_getaffinity(0))
        except AttributeError:  # no cov
            w_num = os.cpu_count() or 1
    for i in range(w_num):
        worker = server_cls(config)
        worker.configure(f"web_worker_{i + 1}", ext_router=router, sock=sock,
                         api_addons_loader=api_addons_loader, **kwargs)
        workers.append(worker)
    return workers


def create_rt_subscribes(*workers: HttpServer,
                         event_type: List[Type[IMessage] | str] | Type[IMessage] | str,
                         resolve_callback: RtEventCallback,
                         **kwargs):
    for worker in workers:
        worker.channel.add_event_listener(event_type, partial(worker.broadcasts_rt, resolve_callback=resolve_callback),
                                          **kwargs)



from web_foundation.kernel.isolate import *
from web_foundation.kernel.messaging.channel import *

__all__ = ("Isolate",
           "IChannel",
           "IMessage",
           "IsolatePipes"
           )



from typing import Dict

import loguru

from web_foundation import settings
from web_foundation.environment.events.metrics import MetricRequest, NewMetricEvent, MetricResponse
from web_foundation.environment.metrics.basemetric import CounterBaseMetric
from web_foundation.environment.metrics.exporter import BaseMetric
from web_foundation.kernel import GenericIMessage
from web_foundation.kernel.messaging.dispatcher import IDispatcher

if settings.METRICS_ENABLE:
    try:
        from prometheus_client.metrics import Gauge, Histogram, Counter, MetricWrapperBase
    except ImportError:
        settings.PROMETHEUS_METRICS_ENABLE = False


class MetricsDispatcher(IDispatcher):
    collected_metrics: Dict[str, Dict[str, BaseMetric]]

    def __init__(self):
        super().__init__()
        self.collected_metrics = {}
        self.add_event_listener(NewMetricEvent.message_type, self.on_new_metric)
        self.add_event_listener(MetricRequest.message_type, self.on_metric_request)

    async def on_metric_request(self, event: MetricRequest):
        exporter = event.exporter
        sender_channel = self.channels.get(event.sender)
        if sender_channel:
            if settings.DEBUG:
                loguru.logger.debug(f"Send METRICS to {sender_channel}")
            if not self.collected_metrics:
                await sender_channel.sent_to_consume(exporter.empty())
            rp = []
            for i in self.collected_metrics.values():
                if isinstance(i, dict):
                    for k in i.values():
                        rp.append(k)
                else:
                    rp.append(i)
            resp = MetricResponse(exporter.export(rp))
            resp.inner_index = event.inner_index
            await sender_channel.sent_to_consume(resp)

    async def on_new_metric(self, event: NewMetricEvent):
        metr = self.collected_metrics.get(event.metric.name)
        if not metr:
            self.collected_metrics.update({event.metric.name: {event.metric.labels_concat: event.metric}})
        else:
            if metr.get(event.metric.labels_concat):
                # loguru.logger.warning(f"get {event.metric}")
                metr[event.metric.labels_concat] += event.metric
            else:
                # loguru.logger.warning(f"set {event.metric}")
                metr[event.metric.labels_concat] = event.metric

    async def track_event(self, msg: GenericIMessage):
        await super(MetricsDispatcher, self).track_event(msg)
        if settings.EVENTS_METRIC_ENABLE:
            event_counter = CounterBaseMetric("events_counter", self._msg_global_index)
            self.collected_metrics.update({"events_counter": {"dispatcher_events_counter": event_counter}})

            named_event_counter = self.collected_metrics.get("named_events_counter")
            if not named_event_counter:
                mtr = CounterBaseMetric("named_events_counter", value=1)
                mtr.add_label(event_name=msg.message_type)
                named_event_counter = {msg.message_type: mtr}
                self.collected_metrics["named_events_counter"] = named_event_counter
            elif named_event_counter.get(msg.message_type):
                named_event_counter[msg.message_type].inc()
            else:
                mtr = CounterBaseMetric("named_events_counter", value=1)
                mtr.add_label(event_name=msg.message_type)
                named_event_counter[msg.message_type] = mtr
            if settings.DEBUG:
                loguru.logger.debug(f'MetricsDispatcher - track event {msg}')



from dataclasses import dataclass
from datetime import datetime
from typing import Callable, Awaitable, Any
from typing import List, TypeVar, Tuple, Generic

from pydantic import BaseModel as PdModel
from tortoise.exceptions import IntegrityError, FieldError
from tortoise.queryset import QuerySet, MODEL

from web_foundation.environment.resources.database.models import ConcurrentEntity
from web_foundation.environment.resources.database.models import GenericDbType
from web_foundation.environment.resources.database.utils import integrity_error_format
from web_foundation.environment.workers.web.ext.request_handler import ProtectIdentity
from web_foundation.errors.app.application import InconsistencyError

EntityId = TypeVar("EntityId", bound=int)


async def retry_dummy(dummy: Any) -> bool:
    return False


@dataclass
class AccessProtectIdentity(Generic[GenericDbType], ProtectIdentity):
    pass


class ModelLoader(Generic[GenericDbType]):

    @classmethod
    async def fetch_related(cls, model: GenericDbType, query: QuerySet[MODEL],
                            fetch_fields: List[str] | None = None) -> QuerySet[MODEL]:
        if fetch_fields:
            for_select = {field for fk_field in model._meta.fk_fields for field in fetch_fields if
                          fk_field == field}
            for_prefetch = set(fetch_fields) - for_select
            if for_select:
                query = query.select_related(*for_select)
            if for_prefetch:
                query = query.prefetch_related(*for_prefetch)
        return query

    @staticmethod
    async def get_entity(model: GenericDbType,
                         entity_id: EntityId) -> GenericDbType:  # TODO TYPED USER
        entity = await model.get_or_none(id=entity_id)
        if not entity:
            raise InconsistencyError(message=f"{model.__name__} not found")  # type: ignore
        return entity

    @staticmethod
    async def create(model: GenericDbType, protection: AccessProtectIdentity, dto: PdModel,
                     **kwargs) -> GenericDbType:
        entity_kwargs = {}
        m2m_fields = {}
        for field, value in dto.dict().items():
            if value is None:
                continue
            if field.endswith('_id') or field.endswith('_ids'):
                raw_field = field[:field.find("_id")]
                if raw_field in model._meta.m2m_fields:
                    m2m_model = model._meta.fields_map[raw_field].model
                    if not isinstance(value, (list, set)):
                        value = [value]
                    m2m_entities = await m2m_model.filter(id__in=value)
                    if len(m2m_entities) != len(value):
                        raise InconsistencyError(message=f"Some {m2m_model.__name__} not found")
                    m2m_fields[raw_field] = m2m_entities
                else:
                    entity_kwargs[field] = value
            else:
                entity_kwargs[field] = value
        try:
            entity = await model.create(**entity_kwargs)
        except IntegrityError as exception:
            raise integrity_error_format(exception)
        except ValueError as exception:
            raise InconsistencyError(exception)

        for m2m_field, m2m_entities in m2m_fields.items():
            await getattr(entity, m2m_field).add(*m2m_entities)
        return entity

    @staticmethod
    async def read(model: GenericDbType, entity_id: EntityId, protection: AccessProtectIdentity,
                   fetch_fields: List[str] | None = None,
                   **kwargs) -> GenericDbType | None:
        query = model.filter(id=entity_id)
        if kwargs:
            query = query.filter(**kwargs)
        query = await ModelLoader.fetch_related(model, query, fetch_fields)
        entity = await query.first()
        return entity

    @staticmethod
    async def read_all(model: GenericDbType, protection: AccessProtectIdentity, limit: int = None, offset: int = None,
                       fetch_fields: List[str] = None,
                       **kwargs) -> Tuple[List[GenericDbType], int]:
        query = model.filter()
        if kwargs:
            if "order_by" in kwargs:
                query = query.order_by(kwargs.pop("order_by"))
            query = query.filter(**kwargs)

        if limit:
            query = query.limit(limit)
        if offset:
            query = query.offset(offset)
        query = await ModelLoader.fetch_related(model, query, fetch_fields)
        try:
            entities = await query
            query._limit = None
            query._offset = None
            total = await query.count()
        except FieldError:
            raise InconsistencyError(message="Incorrect filter")
        return entities, total

    @staticmethod
    async def update(model: GenericDbType, entity_id: EntityId, protection: AccessProtectIdentity, dto: PdModel,
                     **kwargs) -> GenericDbType:
        entity = await ModelLoader.get_entity(model, entity_id)
        m2m_fields = {}
        entity_kwargs = {field: value for field, value in dto.dict().items() if value is not None}
        for field, value in entity_kwargs.items():
            if field.endswith('_id') or field.endswith('_ids'):
                raw_field = field[:field.find("_id")]
                if raw_field in model._meta.m2m_fields:
                    m2m_model = model._meta.fields_map[raw_field].model
                    if not isinstance(value, (list, set)):
                        value = [value]
                    m2m_entities = await m2m_model.filter(id__in=value)
                    if len(m2m_entities) != len(value):
                        raise InconsistencyError(message=f"Some {m2m_model.__name__} not found")
                    m2m_fields[raw_field] = m2m_entities
                    continue
            setattr(entity, field, value)
        try:
            await entity.save(update_fields=list(entity_kwargs.keys()))
        except IntegrityError as exception:
            raise integrity_error_format(exception)
        except ValueError as exception:
            raise InconsistencyError(exception)
        for m2m_field, m2m_entities in m2m_fields.items():
            await getattr(entity, m2m_field).add(*m2m_entities)
        return entity

    @staticmethod
    async def delete(model: GenericDbType, entity_id: EntityId, protection: AccessProtectIdentity,
                     **kwargs) -> GenericDbType:
        assert entity_id is not None
        entity = await ModelLoader.get_entity(model, entity_id)
        await entity.delete()
        return entity

    @classmethod
    async def concurrent_update(cls, query: QuerySet, should_retry: Callable[[Any], Awaitable[bool]] = retry_dummy,
                                **kwargs):
        """
            This function is designed to apply updates to database objects in concurrent manner
            and uses Read-Modify-Write Pattern (RMW).

            It's implemented in three basic steps:
            1. Read object -> acquire/own it;
            2. Modify local copy;
            3. Apply updates only if we still own it (we have it's latest version)

            Step 3 may fail in case when concurrency detected:
            some other database session modified target object
            while we were modifying our local copy, which means
            we do not 'own' the object and must refresh it
            in order to get the latest version.

            It is perceived by comparing local 'updated_at' field
            with it's actual value at step 3.
            So the typical produced query is going to look like:
            update table set value=x where updated_at=<our local updated_at value>

            'query' is required to contain some filtering expression
            'apply' must contain kwargs to be applied in case of success
            'should_retry' functor can be provided as guard who checks retry neccesarity

            NOTE: To keep database consistency regarding
            concurrently accessed objects - every code must MODIFY
            such objects only through this function.

            NOTE: requires connections to use 'read committed' policy
        """
        while True:
            entity: ConcurrentEntity = await query.first()
            if not entity:
                return None

            kwargs.update({'updated_at': datetime.now()})
            done = await query.filter(updated_at=entity.updated_at) \
                .update(**kwargs)

            if done:
                return await query.first()
            elif not await should_retry(query):
                return None
